module Lexer = struct
  include Token

  type lexer = {
    input : string;
    position : int;
    read_position : int;
    ch : char;
  }

  let null_byte = '\x00'


  let advance lexer =
    let read_to_end = lexer.read_position >= String.length(lexer.input) in
    let new_ch = if read_to_end then null_byte else String.get lexer.input lexer.read_position in
    let new_lexer = {lexer with position = lexer.read_position; read_position = lexer.read_position + 1; ch = new_ch} in
    new_lexer

  let peek_char lexer =
    let read_to_end = lexer.read_position >= String.length(lexer.input) - 1 in
    if read_to_end then null_byte else String.get lexer.input (lexer.read_position)
  
  let rec advance_to ch lexer =
    if peek_char lexer != ch then advance lexer |> advance_to ch
    else advance lexer

  let rec advance_to2 ch nch lexer =
    let _ = print_endline (Format.sprintf "%c%c\n" ch nch) in
    let lex = advance_to ch lexer in
    let _ = print_endline (Format.sprintf "%c\n" (peek_char lex)) in
    if peek_char lex = nch then lex else advance_to2 ch nch lex


  let rec skip_whitespace lexer = 
    match lexer.ch with
    | ' ' | '\n' | '\t' -> advance lexer |> skip_whitespace
    | _ -> lexer
    

  let new_lexer input_string =
    let lexer = {
      input = input_string;
      position = 0;
      read_position = 0;
      ch = null_byte;
    } in
    advance lexer

  let read_string lexer =
    let rec read_str lxr str =
      match lxr.ch with
      | '"' -> (advance lxr, Token.STRING str)
      | c -> read_str (advance lxr) (str ^ Char.escaped c) in
    read_str lexer ""

  let rec read_ident lxr str = 
    match lxr.ch with
    | '0'..'9' | 'a'..'z' | 'A'..'Z' | '_' as c -> read_ident (advance lxr) (str ^ Char.escaped c)
    | _ -> match str with
      | "null" -> (lxr, Token.NULL)
      | "true" | "false" -> (lxr, Token.BOOL (bool_of_string str))
      | _ -> (lxr, Token.IDENT str)

  let read_number lexer =
    let rec read_decimals lxr str = 
      match lxr.ch with
      | '0'..'9' as c -> read_decimals (advance lxr) (str ^ Char.escaped c)
      | 'f' -> (advance lxr, Token.FLOAT (float_of_string str)) 
      | _ -> (lxr, Token.FLOAT (float_of_string str))
    in
    let rec read_num lxr str =
      match lxr.ch with
      | '0'..'9' as c -> read_num (advance lxr) (str ^ Char.escaped c)
      | 'f' -> (advance lxr, Token.FLOAT (float_of_string str))
      | '.' -> read_decimals (advance lxr) (str ^ Char.escaped '.')
      | _ -> (lxr, Token.INTEGER (int_of_string str))
    in
    read_num lexer ""

  let read_binary lexer =
    let rec read_bin lxr str =
      match lxr.ch with
      | '"' -> (advance lxr, Token.BINARY (Bytes.of_string str))
      | c -> read_bin (advance lxr) (str ^ Char.escaped c) in
    read_bin lexer ""

  let rec next_token lexer = 
    let lxr = skip_whitespace lexer in
    match lxr.ch with
    | ':' -> (advance lxr, Token.COLON)
    | ',' -> (advance lxr, Token.COMMA)
    | '{' -> (advance lxr, Token.LBRACE)
    | '}' -> (advance lxr, Token.RBRACE)
    | '[' -> (advance lxr, Token.LBRACKET)
    | ']' -> (advance lxr, Token.RBRACKET)
    | '/' when peek_char lxr = '/' -> let l = advance_to '\n' lxr in next_token l
    | '/' when peek_char lxr = '*' -> let l = advance lxr |> advance |> advance_to2 '*' '/' in advance l |> advance |> next_token
    | '"' -> advance lxr |> read_string
    | 'b' when peek_char lxr = '"' -> advance lxr |> advance |> read_binary
    | 'a'..'z' | 'A'..'Z' as c -> read_ident (advance lxr) (Char.escaped c)
    | '0'..'9' -> read_number lxr
    | '\x00' -> (lxr, Token.EOF)
    | _ -> (advance lxr, Token.ILLEGAL)

  let generate_tokens input_string =
    let lexer = new_lexer input_string in
    let rec gen lxr tokens =
      match next_token lxr with
      | (_, Token.EOF) -> List.rev_append tokens [Token.EOF]
      | (l, tok) -> gen l (tok :: tokens)
    in gen lexer []
end